#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "./list.h"

struct Node {
  int val;
  Node* prev;
  Node* next;
};

Node* node_next(Node* n) {
  return n->next;
}
Node* node_prev(Node* n) {
  return n->prev;
}
int* node_val(Node* n) {
  return &n->val;
}

struct List {
  Node* first;
  Node* last;
  size_t len;
};

List* list_new(void) {
  List* l = malloc(sizeof(List));
  if (l == NULL) {
    exit(1);
  }
  l->first = NULL;
  l->last = NULL;
  l->len = 0;
  return l;
}

void list_free(List* l) {
  Node* n = l->first;
  while (n != NULL) {
    Node* tmp = n->next;
    free(n);
    n = tmp;
  }
  free(l);
}

Node* list_first(List* l) {
  return l->first;
}
Node* list_last(List* l) {
  return l->last;
}
size_t list_len(List* l) {
  return l->len;
}

void list_push_back(List* l, int val) {
  Node* n = malloc(sizeof(Node));
  if (n == NULL) {
    exit(1);
  }
  n->val = val;
  n->prev = l->last;
  n->next = NULL;

  if (l->last == NULL) {
    l->first = n;
  } else {
    l->last->next = n;
  }
  l->last = n;
  l->len++;
}

void list_push_front(List* l, int val) {
  Node* n = malloc(sizeof(Node));
  if (n == NULL) {
    exit(1);
  }
  n->val = val;
  n->prev = NULL;
  n->next = l->first;

  if (l->first == NULL) {
    l->last = n;
  } else {
    l->first->prev = n;
  }
  l->first = n;
  l->len++;
}

int list_pop_back(List* l) {
  if (l->first == l->last) {
    int val = l->last->val;
    free(l->first);
    l->first = NULL;
    l->last = NULL;
    l->len--;
    return val;
  } else {
    int val = l->last->val;
    Node* last = l->last;
    l->last = l->last->prev;
    free(last);
    l->last->next = NULL;
    l->len--;
    return val;
  }
}

int list_pop_front(List* l) {
  if (l->first == l->last) {
    int val = l->last->val;
    free(l->first);
    l->first = NULL;
    l->last = NULL;
    l->len--;
    return val;
  } else {
    int val = l->first->val;
    Node* first = l->first;
    l->first = l->first->next;
    free(first);
    l->first->prev = NULL;
    l->len--;
    return val;
  }
}
