#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>

#include "./table_lib.h"

char* fgetline(FILE* file) {
  char* line = NULL;
  size_t len = 0;
  ssize_t success = getline(&line, &len, file);
  if (success <= 0) {
    free(line);
    return NULL;
  } else {
    return line;
  }
}

FILE** open_files(char** file_paths, size_t num_file_paths, FILE* error_file) {
  FILE** files = malloc(num_file_paths * sizeof(FILE*));
  if (files == NULL) {
    return NULL;
  }
  bool error = false;
  for (size_t i = 0; i < num_file_paths; ++i) {
    FILE* f = fopen(file_paths[i], "r");
    if (f == NULL) {
      fprintf(error_file, "ERROR: Failed to open file '%s'.\n", file_paths[i]);
      error = true;
    }
    files[i] = f;
  }
  if (error) {
    for (size_t i = 0; i < num_file_paths; ++i) {
      if (files[i] != NULL) {
        fclose(files[i]);
      }
    }
    free(files);
    return NULL;
  }
  return files;
}

void close_files(FILE** files, size_t num_files) {
  for (size_t i = 0; i < num_files; ++i) {
    fclose(files[i]);
  }
  free(files);
}

void merge_columns(FILE** input_files, size_t num_input_files, FILE* output_file) {
  char** lines = malloc(sizeof(char*) * num_input_files);

  while (true) {
    /* Try to read a line from each file. */
    bool done_with_all_files = true;
    for (size_t i = 0; i < num_input_files; ++i) {
      lines[i] = fgetline(input_files[i]);
      done_with_all_files = done_with_all_files && lines[i] == NULL;
    }

    /* If all files are read to completion, we're done. */
    if (done_with_all_files) {
      free(lines);
      return;
    }

    /* Otherwise, we print a row with the read lines as cells. */
    for (size_t i = 0; i < num_input_files; ++i) {
      /* If the file wasn't read to completion, we print the cell content. */
      if (lines[i] != NULL) {
        /* First we remove trailing '\n' character of the read line. */
        for (char* p = lines[i]; *p != 0; ++p) {
          if (*p == '\n') {
            *p = 0;
            break;
          }
        }

        /* Then we print the read line */
        fprintf(output_file, "%s", lines[i]);

        free(lines[i]);
      }
      if (i < num_input_files - 1) {
        fprintf(output_file, ",");
      } else {
        fprintf(output_file, "\n");
      }
    }
  }
}
